import boto3.s3.transfer

from ..file import File

import os
import hashlib
from typing import Optional, Union

KB = 1024
MB = KB * KB
TransferConfig = boto3.s3.transfer.TransferConfig(
    max_concurrency=1, multipart_chunksize=8 * MB, multipart_threshold=8 * MB
)


def calculate_etag(file: Union[str, os.PathLike]):
    # taken from https://stackoverflow.com/a/52300584
    with open(file, "rb") as f:
        md5hash = hashlib.md5()
        filesize = 0
        block_count = 0
        md5string = b""
        for block in iter(
            lambda: f.read(TransferConfig.multipart_chunksize), b""
        ):
            md5hash = hashlib.md5()
            md5hash.update(block)
            md5string += md5hash.digest()
            filesize += len(block)
            block_count += 1

    if filesize > TransferConfig.multipart_threshold:
        md5hash = hashlib.md5()
        md5hash.update(md5string)
        md5hash_digested = md5hash.hexdigest() + "-" + str(block_count)
    else:
        md5hash_digested = md5hash.hexdigest()

    return md5hash_digested


# taken from https://boto3.amazonaws.com/v1/documentation/api/latest/guide/s3-uploading-files.html
class S3ProgressPercentage(object):
    def __init__(
        self,
        file: File,
        filename: Union[str, os.PathLike],
        operation_index: int,
        size: Optional[float] = None,
    ):
        self._file = file
        if size:
            self._size = size
        else:
            self._size = float(os.path.getsize(filename))
        self._seen_so_far = 0
        self._last_percentage = 0
        self._operation_index = operation_index

    def __call__(self, bytes_amount):
        # To simplify, assume this is hooked up to a single filename
        self._seen_so_far += bytes_amount
        percentage = (self._seen_so_far / self._size) * 100
        if int(percentage) > self._last_percentage:
            self._last_percentage = int(percentage)
            self._file.update_progressbar(
                self._operation_index, self._last_percentage
            )



from __future__ import annotations

import requests

from . import ExitableThread
from ..version import __version__

import logging
from queue import Queue, Full
import time
import uuid
from pathlib import Path
from typing import Optional

logger = logging.getLogger(__name__)


class GoogleAnalyticsConsumer(ExitableThread):
    def __init__(self, context):
        super().__init__()
        self.daemon = True
        self._context = context
        self._session = requests.Session()

    def run(self):
        while not self.should_exit:
            if not self._context._queue.empty():
                ec, ea, el = self._context._queue.get()
                data = self._context._base_payload.copy()
                data.update(ec=ec, ea=ea)
                if el:
                    data.update(el=el)
                response = self._session.post(
                    url=self._context._endpoint,
                    params=data,
                    allow_redirects=True,
                )
                try:
                    response.raise_for_status()
                except Exception as e:
                    pass

            time.sleep(0.1)


class GoogleAnalyticsContext:
    def __init__(
        self,
        endpoint: str,
        tracking_id: str,
        application_name: str,
        application_version: str,
        config_file: Path,
    ):
        self._endpoint = endpoint
        self._tracking_id = tracking_id
        self._application_name = application_name
        self._application_version = application_version
        self._config_file = config_file

        self._queue: Queue = Queue(maxsize=10000)

        if self._config_file.exists() and self._config_file.is_file():
            _uuid = self._config_file.read_text().strip()
            if not self._valid_uuid(_uuid):
                _uuid = str(uuid.uuid4())
                self._config_file.write_text(_uuid)
        else:
            _uuid = str(uuid.uuid4())
            try:
                self._config_file.parent.mkdir(
                    mode=0o700, parents=True, exist_ok=True
                )
                self._config_file.write_text(_uuid)
            except Exception:
                pass

        self._base_payload = dict(
            v=1,  # protocol version
            tid=tracking_id,
            t="event",
            an=application_name,
            av=application_version,
            cid=_uuid,
            # without user-agent, requests does not work!
            ua="Opera/9.80 (Windows NT 6.0) Presto/2.12.388 Version/12.14",
        )

        # launch consumer thread
        self._consumer_thread = GoogleAnalyticsConsumer(self)
        self._consumer_thread.start()

    def __del__(self):
        if self._consumer_thread.is_alive():
            self._consumer_thread.should_exit = True

    @property
    def consumer_thread(self):
        return self._consumer_thread

    def send_event(
        self, category: str, action: str, label: Optional[str] = None
    ):
        if len(category) > 150:
            logger.warning(f"Category {category} is too long")
            return
        elif len(action) > 500:
            logger.warning(f"Action {action} is too long")
            return
        elif label is not None and len(label) > 500:
            logger.warning(f"Label {label} is too long")
            return

        try:
            if self._consumer_thread.is_alive():
                self._queue.put_nowait((category, action, label))
        except Full:
            logger.exception("queue is full!")

    def _valid_uuid(self, _uuid):
        try:
            uuid.UUID(_uuid)
        except ValueError:
            return False
        return True



from __future__ import annotations

import gi

gi.require_version("Gtk", "3.0")
from gi.repository import Gio, GLib, Gtk
from tenacity.retry import retry_base
from requests.adapters import HTTPAdapter

from typing import (
    Callable,
    Optional,
    Final,
    Any,
    Dict,
    List,
    Iterable,
    Union,
    Set,
)
import logging
from pathlib import Path, PurePath
import hashlib
import os
import platform
from threading import Thread, current_thread
import time
import string
import random

from .exceptions import SkippedOperation

# bump this number when the yaml layout changes!
MONITOR_YAML_VERSION = 2

EXPAND_AND_FILL: Final[Dict[str, Any]] = dict(
    hexpand=True, vexpand=True, halign=Gtk.Align.FILL, valign=Gtk.Align.FILL
)

PREFERENCES_CONFIG_FILE = Path(
    GLib.get_user_config_dir(), "rfi-file-monitor", "prefs.yml"
)

PATTERN_PLACEHOLDER_TEXT = "e.g *.txt, *.csv or *temp* or *log*"

DEFAULT_TIMEOUT = 5  # seconds

logger = logging.getLogger(__name__)


class TimeoutHTTPAdapter(HTTPAdapter):
    def __init__(self, *args, **kwargs):
        self.timeout = DEFAULT_TIMEOUT
        if "timeout" in kwargs:
            self.timeout = kwargs["timeout"]
            del kwargs["timeout"]
        super().__init__(*args, **kwargs)

    def send(self, request, **kwargs):
        timeout = kwargs.get("timeout")
        if timeout is None:
            kwargs["timeout"] = self.timeout
        return super().send(request, **kwargs)


# had to write my own retry condition class to support:
# 1. retry if an exception was thrown that was not a SkippedOperation
# 2. retry if a value was returned that is not None
# see https://github.com/jd/tenacity/issues/255
class monitor_retry_condition(retry_base):
    def __call__(self, retry_state):
        if retry_state.outcome.failed:
            return self._exception_predicate(retry_state.outcome.exception())
        else:
            return self._value_predicate(retry_state.outcome.result())

    # the following two methods will first check if the thread they are running in has been marked for exiting
    @staticmethod
    def _value_predicate(value):
        if (
            isinstance(current_thread(), ExitableThread)
            and current_thread().should_exit is True
        ):
            return False
        return value is not None

    @staticmethod
    def _exception_predicate(e):
        if (
            isinstance(current_thread(), ExitableThread)
            and current_thread().should_exit is True
        ):
            return False
        return not isinstance(e, SkippedOperation)


def query_metadata(
    metadata: Dict[int, Dict[str, Any]], key: str, full_dict=False
) -> Any:
    """
    Reverse iterates through the metadata until the key is found.
    When successful, returns the matching dictionary value.
    If full_dict is True, then the whole dict that the key belongs to is returned.
    Upon failure, None is returned.
    """
    if metadata is None:
        return None
    for metadata_dict in reversed(metadata.values()):
        if key in metadata_dict:
            if full_dict:
                return metadata_dict
            else:
                return metadata_dict[key]
    return None


def add_action_entries(
    map: Gio.ActionMap,
    action: str,
    callback: Callable[[Gio.ActionMap, Gio.SimpleAction, GLib.Variant], None],
    param: Optional[str] = None,
    state: Optional[GLib.Variant] = None,
    callback_arg: Optional[Any] = None,
) -> None:

    if state:
        simple_action = Gio.SimpleAction.new_stateful(
            action, GLib.VariantType.new(param) if param else None, state
        )
    else:
        simple_action = Gio.SimpleAction.new(
            action, GLib.VariantType.new(param) if param else None
        )

    if callback_arg:
        simple_action.connect("activate", callback, callback_arg)
    else:
        simple_action.connect("activate", callback)

    map.add_action(simple_action)


def class_in_object_iterable(iterable: Iterable, klass) -> bool:
    for _iter in iterable:
        if isinstance(_iter, klass):
            return True
    return False


def get_patterns_from_string(
    input: str, defaults: Optional[Iterable[str]] = None
) -> List[str]:
    if defaults is None:
        if input and input.strip():
            return list(map(lambda x: x.strip(), input.split(",")))
        else:
            return ["*"]
    else:
        if input and input.strip():
            rv = list(map(lambda x: x.strip(), input.split(",")))
            rv.extend(defaults)
            return rv
        else:
            return list(defaults)


def get_md5(fname: Union[os.PathLike, str]) -> str:
    # taken from https://stackoverflow.com/a/3431838
    hash_md5 = hashlib.md5()
    with open(fname, "rb") as f:
        for chunk in iter(lambda: f.read(4096), b""):
            hash_md5.update(chunk)
    return hash_md5.hexdigest()


def get_file_creation_timestamp(file_path: Union[os.PathLike, str]):
    # get creation time, or something similar...
    # https://stackoverflow.com/a/39501288
    if platform.system() == "Windows":
        try:
            return os.stat(file_path).st_ctime
        except FileNotFoundError:
            time.sleep(1)
            try:
                return os.stat(file_path).st_ctime
            except FileNotFoundError:
                return None
    else:
        try:
            # this should work on macOS
            return os.stat(file_path).st_birthtime
        except AttributeError:
            return os.stat(file_path).st_mtime


def get_random_string(length: int):
    letters = string.ascii_letters
    result_str = "".join(random.choice(letters) for i in range(length))
    return result_str


class LongTaskWindow(Gtk.Window):
    def __init__(
        self, parent_window: Optional[Gtk.Window] = None, *args, **kwargs
    ):
        kwargs.update(
            dict(
                transient_for=parent_window,
                window_position=Gtk.WindowPosition.CENTER_ON_PARENT,
                modal=True,
                default_width=250,
                default_height=100,
                type=Gtk.WindowType.TOPLEVEL,
                destroy_with_parent=True,
                decorated=False,
                border_width=5,
            )
        )
        Gtk.Window.__init__(self, *args, **kwargs)
        main_grid = Gtk.Grid(
            column_spacing=10, row_spacing=10, **EXPAND_AND_FILL
        )
        self._label = Gtk.Label(wrap=True, **EXPAND_AND_FILL)
        main_grid.attach(self._label, 0, 0, 1, 1)
        label = Gtk.Label(
            label="This may take a while...",
        )
        main_grid.attach(label, 0, 1, 1, 1)
        self.add(main_grid)
        self.connect("delete-event", Gtk.true)
        main_grid.show_all()

    def set_text(self, text: str):
        self._label.set_markup(text)


class ExitableThread(Thread):
    def __init__(self):
        super().__init__()
        self._should_exit: bool = False

    @property
    def should_exit(self):
        return self._should_exit

    @should_exit.setter
    def should_exit(self, value: bool):
        self._should_exit = value


def _get_common_patterns(
    included_patterns: List[str],
    excluded_patterns: List[str],
    case_sensitive: bool,
) -> Set[str]:
    if not case_sensitive:
        included_patterns = [x.lower() for x in included_patterns] + [
            x.upper() for x in included_patterns
        ]
        excluded_patterns = [x.lower() for x in excluded_patterns] + [
            x.upper() for x in excluded_patterns
        ]
    return set(included_patterns).intersection(excluded_patterns)


def match_path(
    path: PurePath,
    included_patterns: List[str],
    excluded_patterns: List[str],
    case_sensitive: bool = True,
) -> bool:
    common_patterns = _get_common_patterns(
        included_patterns, excluded_patterns, case_sensitive
    )

    if common_patterns:
        raise ValueError(
            f"conflicting patterns `{common_patterns}` included and excluded"
        )
    return any(path.match(p) for p in included_patterns) and not any(
        path.match(p) for p in excluded_patterns
    )



from __future__ import annotations

import gi

gi.require_version("Gtk", "3.0")
gi.require_version("Gdk", "3.0")
from gi.repository import Gtk, GObject, Gdk, GLib
from munch import Munch

from abc import ABC, abstractmethod, ABCMeta
from typing import Optional, Type, Union
import traceback
import logging

from .utils import ExitableThread, LongTaskWindow
from .utils.exceptions import AlreadyRunning, NotYetRunning
from .utils.widgetparams import WidgetParams


logger = logging.getLogger(__name__)


class EngineThread(ABC, ExitableThread):
    def __init__(self, engine: Engine, task_window: LongTaskWindow):
        super().__init__()
        self._engine: Engine = engine
        self._task_window = task_window
        self.params = engine._get_params()


class EngineMeta(ABCMeta, Gtk.Grid.__class__):
    pass


class Engine(ABC, WidgetParams, Gtk.Grid, metaclass=EngineMeta):
    @abstractmethod
    def __init__(
        self,
        appwindow,
        engine_thread_class: Type[EngineThread],
        abort_message: str,
    ):
        self._appwindow = appwindow
        self._engine_thread_class = engine_thread_class
        self._abort_message = abort_message

        kwargs = dict(
            halign=Gtk.Align.FILL,
            valign=Gtk.Align.FILL,
            hexpand=True,
            vexpand=True,
            row_spacing=5,
            column_spacing=5,
            border_width=5,
        )
        Gtk.Grid.__init__(self, **kwargs)
        WidgetParams.__init__(self)

        self._running = False
        self._valid = False

    @property
    @classmethod
    @abstractmethod
    def NAME(cls) -> str:
        """
        A descriptive short name for the engine
        """
        raise NotImplementedError

    @GObject.Property(type=bool, default=False)
    def running(self):
        return self._running

    @GObject.Property(type=bool, default=False)
    def valid(self):
        return self._valid

    @property
    def appwindow(self):
        return self._appwindow

    def start(self):
        if self._running:
            raise AlreadyRunning(
                "The engine is already running. It needs to be stopped before it may be restarted"
            )

        # pop up a long task window
        # spawn a thread for this to avoid GUI freezes
        task_window = LongTaskWindow(self._appwindow)
        task_window.set_text(f"<b>Launching {self.NAME}</b>")
        task_window.show()
        watch_cursor = Gdk.Cursor.new_for_display(
            Gdk.Display.get_default(), Gdk.CursorType.WATCH
        )
        task_window.get_window().set_cursor(watch_cursor)

        self._thread = self._engine_thread_class(self, task_window)
        self._thread.start()

    def stop(self):
        if not self._running:
            raise NotYetRunning(
                "The engine needs to be started before it can be stopped."
            )

        task_window = LongTaskWindow(self._appwindow)
        task_window.set_text(f"<b>Stopping {self.NAME}</b>")
        task_window.show()
        watch_cursor = Gdk.Cursor.new_for_display(
            Gdk.Display.get_default(), Gdk.CursorType.WATCH
        )
        task_window.get_window().set_cursor(watch_cursor)

        self._running_changed_handler_id = self.connect(
            "notify::running", self._running_changed_cb, task_window
        )

        # if the thread is sleeping, it will be killed at the next iteration
        self._thread.should_exit = True

    def _running_changed_cb(self, _self, param, task_window):
        self.disconnect(self._running_changed_handler_id)
        if not self._running:
            task_window.get_window().set_cursor(None)
            task_window.destroy()

    def kill_task_window(self, task_window: LongTaskWindow):
        # destroy task window
        task_window.get_window().set_cursor(None)
        task_window.destroy()

        self._running = True
        self.notify("running")

        return GLib.SOURCE_REMOVE

    def abort(
        self, task_window: Optional[LongTaskWindow], e: Union[Exception, str]
    ):
        if task_window:
            # destroy task window
            task_window.get_window().set_cursor(None)
            task_window.destroy()

        # display dialog with error message
        if isinstance(e, Exception):
            traceback.print_exc()
            logger.info("".join(traceback.format_tb(e.__traceback__)))
        elif isinstance(e, str):
            logger.info(e)
        dialog = Gtk.MessageDialog(
            transient_for=self.get_toplevel(),
            modal=True,
            destroy_with_parent=True,
            message_type=Gtk.MessageType.ERROR,
            buttons=Gtk.ButtonsType.CLOSE,
            text=f"Could not start {self.NAME}",
            secondary_text=str(e) + f"\n\n{self._abort_message}",
        )
        dialog.run()
        dialog.destroy()

        return GLib.SOURCE_REMOVE

    def cleanup(self):
        GLib.idle_add(self._stop_running)

    def _stop_running(self):
        self._running = False
        self.notify("running")

        return GLib.SOURCE_REMOVE

    def _get_params(self) -> Munch:
        return self.params.copy()



print("import: 'rfi_file_monitor'")
import rfi_file_monitor

print("import: 'rfi_file_monitor.application'")
import rfi_file_monitor.application

print("import: 'rfi_file_monitor.applicationwindow'")
import rfi_file_monitor.applicationwindow

print("import: 'rfi_file_monitor.engine'")
import rfi_file_monitor.engine

print("import: 'rfi_file_monitor.engine_advanced_settings'")
import rfi_file_monitor.engine_advanced_settings

print("import: 'rfi_file_monitor.engines'")
import rfi_file_monitor.engines

print("import: 'rfi_file_monitor.engines.aws_s3_bucket_engine'")
import rfi_file_monitor.engines.aws_s3_bucket_engine

print("import: 'rfi_file_monitor.engines.aws_s3_bucket_engine_advanced_settings'")
import rfi_file_monitor.engines.aws_s3_bucket_engine_advanced_settings

print("import: 'rfi_file_monitor.engines.base_s3_bucket_engine'")
import rfi_file_monitor.engines.base_s3_bucket_engine

print("import: 'rfi_file_monitor.engines.ceph_s3_bucket_engine'")
import rfi_file_monitor.engines.ceph_s3_bucket_engine

print("import: 'rfi_file_monitor.engines.ceph_s3_bucket_engine_advanced_settings'")
import rfi_file_monitor.engines.ceph_s3_bucket_engine_advanced_settings

print("import: 'rfi_file_monitor.engines.directory_watchdog_engine'")
import rfi_file_monitor.engines.directory_watchdog_engine

print("import: 'rfi_file_monitor.engines.directory_watchdog_engine_advanced_settings'")
import rfi_file_monitor.engines.directory_watchdog_engine_advanced_settings

print("import: 'rfi_file_monitor.engines.file_watchdog_engine_advanced_settings'")
import rfi_file_monitor.engines.file_watchdog_engine_advanced_settings

print("import: 'rfi_file_monitor.engines.file_watchdog_engine'")
import rfi_file_monitor.engines.file_watchdog_engine

print("import: 'rfi_file_monitor.engines.temporary_file_engine'")
import rfi_file_monitor.engines.temporary_file_engine

print("import: 'rfi_file_monitor.engines.urls_engine'")
import rfi_file_monitor.engines.urls_engine

print("import: 'rfi_file_monitor.file'")
import rfi_file_monitor.file

print("import: 'rfi_file_monitor.job'")
import rfi_file_monitor.job

print("import: 'rfi_file_monitor.operation'")
import rfi_file_monitor.operation

print("import: 'rfi_file_monitor.operations.directory_compressor'")
import rfi_file_monitor.operations.directory_compressor

print("import: 'rfi_file_monitor.operations.dropbox_uploader'")
import rfi_file_monitor.operations.dropbox_uploader

print("import: 'rfi_file_monitor.operations.dummy_operation'")
import rfi_file_monitor.operations.dummy_operation

print("import: 'rfi_file_monitor.operations.local_copier'")
import rfi_file_monitor.operations.local_copier

print("import: 'rfi_file_monitor.operations.s3_copier'")
import rfi_file_monitor.operations.s3_copier

print("import: 'rfi_file_monitor.operations.s3_downloader'")
import rfi_file_monitor.operations.s3_downloader

print("import: 'rfi_file_monitor.operations.s3_uploader'")
import rfi_file_monitor.operations.s3_uploader

print("import: 'rfi_file_monitor.operations.sftp_uploader'")
import rfi_file_monitor.operations.sftp_uploader

print("import: 'rfi_file_monitor.preferences'")
import rfi_file_monitor.preferences

print("import: 'rfi_file_monitor.preferenceswindow'")
import rfi_file_monitor.preferenceswindow

print("import: 'rfi_file_monitor.queue_manager'")
import rfi_file_monitor.queue_manager

print("import: 'rfi_file_monitor.utils'")
import rfi_file_monitor.utils

print("import: 'rfi_file_monitor.utils.decorators'")
import rfi_file_monitor.utils.decorators

print("import: 'rfi_file_monitor.utils.exceptions'")
import rfi_file_monitor.utils.exceptions

print("import: 'rfi_file_monitor.utils.googleanalytics'")
import rfi_file_monitor.utils.googleanalytics

print("import: 'rfi_file_monitor.utils.helpwindow'")
import rfi_file_monitor.utils.helpwindow

print("import: 'rfi_file_monitor.utils.paramswindow'")
import rfi_file_monitor.utils.paramswindow

print("import: 'rfi_file_monitor.utils.s3'")
import rfi_file_monitor.utils.s3

print("import: 'rfi_file_monitor.utils.widgetparams'")
import rfi_file_monitor.utils.widgetparams

print("import: 'rfi_file_monitor.version'")
import rfi_file_monitor.version




import gi

gi.require_version("Gtk", "3.0")
from gi.repository import Gtk, Pango

import logging
from typing import Dict, Final, Type, Union

from . import EXPAND_AND_FILL
from ..engine import Engine
from ..operation import Operation
from ..queue_manager import QueueManager

logger = logging.getLogger(__name__)


class HelpWindow(Gtk.Window):
    def __init__(
        self,
        contents_raw: Final[
            Dict[Type[Union[Engine, QueueManager, Operation]], str]
        ],
    ):
        super().__init__(
            destroy_with_parent=True,
            window_position=Gtk.WindowPosition.NONE,
            border_width=5,
            title="Help",
        )
        self.set_default_size(600, 600)

        grid = Gtk.Grid(**EXPAND_AND_FILL, row_spacing=5, column_spacing=5)
        self.add(grid)

        self._search_entry = Gtk.SearchEntry(
            halign=Gtk.Align.START,
            valign=Gtk.Align.CENTER,
            hexpand=False,
            vexpand=False,
        )
        self._search_entry.connect("changed", self._search_entry_changed_cb)
        grid.attach(self._search_entry, 0, 0, 1, 1)

        list_sw = Gtk.ScrolledWindow(
            halign=Gtk.Align.FILL,
            valign=Gtk.Align.FILL,
            hexpand=False,
            vexpand=True,
            hscrollbar_policy=Gtk.PolicyType.NEVER,
            shadow_type=Gtk.ShadowType.IN,
        )
        grid.attach(list_sw, 0, 1, 1, 1)

        self._list_box = Gtk.ListBox(
            **EXPAND_AND_FILL,
            activate_on_single_click=False,
        )

        list_sw.add(self._list_box)

        # prepare contents
        contents = list()

        # engines
        engines = list(
            sorted(
                filter(
                    lambda widgetclass: issubclass(widgetclass, Engine),
                    contents_raw,
                ),
                key=lambda engine: engine.NAME,
            )
        )
        for engine in engines:
            engine.section = "Engines"
        contents.extend(engines)

        # operations
        operations = list(
            sorted(
                filter(
                    lambda widgetclass: issubclass(widgetclass, Operation),
                    contents_raw,
                ),
                key=lambda operation: operation.NAME,
            )
        )
        for operation in operations:
            operation.section = "Operations"
        contents.extend(operations)

        # others
        others = list(
            sorted(
                filter(
                    lambda widgetclass: not issubclass(widgetclass, Engine)
                    and not issubclass(widgetclass, Operation),
                    contents_raw,
                ),
                key=lambda other: other.NAME,
            )
        )
        for other in others:
            other.section = "Miscellaneous"
        contents.extend(others)

        self._rows = dict()

        self._list_box.set_header_func(self._list_box_header_func)
        self._list_box.set_filter_func(self._list_box_filter_func)

        for widgetclass in contents:
            label = Gtk.Label(
                halign=Gtk.Align.START,
                valign=Gtk.Align.CENTER,
                hexpand=False,
                vexpand=False,
                margin=6,
                label=f"{widgetclass.NAME}",
            )
            self._list_box.insert(label, -1)

            row = label.get_parent()
            row.set_activatable(False)

            self._rows[widgetclass] = row

            row.section = widgetclass.section
            row.label = contents_raw[widgetclass]

        self._list_box.connect("row-selected", self._row_selected_cb)

        label_sw = Gtk.ScrolledWindow(
            **EXPAND_AND_FILL,
            hscrollbar_policy=Gtk.PolicyType.NEVER,
            shadow_type=Gtk.ShadowType.IN,
        )
        self._contents_label = Gtk.Label(
            **EXPAND_AND_FILL,
            wrap=True,
            use_markup=True,
            selectable=True,
            wrap_mode=Pango.WrapMode.WORD_CHAR,
            xalign=0,
            yalign=0,
            margin_start=5,
            margin_end=5,
            margin_top=5,
            margin_bottom=5,
        )
        label_sw.add(self._contents_label)
        grid.attach(label_sw, 1, 0, 1, 2)

        self._list_box.invalidate_headers()

        grid.show_all()

    def _search_entry_changed_cb(self, entry):
        self._list_box.invalidate_filter()

    def _list_box_filter_func(self, row):
        search_string = self._search_entry.get_text().strip()

        if not search_string:
            return True

        return search_string.lower() in row.get_child().props.label.lower()

    def _list_box_header_func(self, row, before):
        header = row.get_header()

        if not header and (before is None or before.section != row.section):
            title = f"<b>{row.section}</b>"

            header = Gtk.Label(
                label=title,
                use_markup=True,
                halign=Gtk.Align.START,
                margin_top=12,
                margin_start=6,
                margin_end=6,
                margin_bottom=6,
            )
            header.show()

            row.set_header(header)

    def _row_selected_cb(self, list_box, row):
        self._contents_label.props.label = row.label

    def select_item(self, klass: Type[Union[Engine, QueueManager, Operation]]):
        self._list_box.select_row(self._rows[klass])

    def do_delete_event(self, event):
        return self.hide_on_delete()



import logging
from time import time

from .file import File, FileStatus
from .utils.exceptions import SkippedOperation
from .utils import ExitableThread

from tenacity import RetryError

logger = logging.getLogger(__name__)


class Job(ExitableThread):

    SKIPPED_MESSAGE = "A preceding operation has been skipped"
    ERROR_MESSAGE = "Operation not started due to previous error"

    def __init__(self, queue_manager, file: File):
        super().__init__()
        self._queue_manager = queue_manager
        self._file = file

    def run(self):
        # update status to running
        self._file.update_status(-1, FileStatus.RUNNING)

        # If operation.run() returns None, then it was considered a success.
        # Otherwise a string is returned with an error message
        rv = None
        # this will contain the first error message we run into,
        # and will be used as tooltip for the parent row
        global_rv = None

        for index, operation in enumerate(
            self._queue_manager._appwindow._operations_box
        ):
            self._file.update_status(index, FileStatus.RUNNING)

            if self._should_exit:
                rv = "Monitoring aborted"
            elif rv is None:
                try:
                    self._queue_manager._appwindow.props.application.google_analytics_context.send_event(
                        "RUN-OPERATION", operation.NAME
                    )
                    rv = operation.run(self._file)
                except SkippedOperation as e:
                    rv = e
                except RetryError as e:
                    # happens when the run method is wrapped with tenacity.retry
                    # and we ran out of retries
                    rv = str(e.last_attempt.result())
                except Exception as e:
                    # exceptions caught here indicate a programming error,
                    # as exceptions should be caught during run, and if necessary,
                    # run should appropriate error message instead of None
                    # The only reason to catch it here is to avoid it taking the app down...
                    rv = str(e)
                    logger.exception(f"run() exception caught: {rv}!")
            else:
                # If we get here then an error was returned in a previous operation already
                rv = self.ERROR_MESSAGE

            if rv is None:
                # update operation status to success
                self._file.update_status(index, FileStatus.SUCCESS)
            elif isinstance(rv, SkippedOperation):
                # update operation status to skipped
                self._file.update_status(index, FileStatus.SKIPPED, str(rv))
                # reset rv to None to ensure the other operations are run
                rv = None
            else:
                # update operation status to failed
                self._file.update_status(index, FileStatus.FAILURE, rv)
                if global_rv is None:
                    global_rv = rv

        # update global operation status
        if global_rv is None:
            # update job status to success
            self._file.update_status(-1, FileStatus.SUCCESS)
            self._file.succeeded = time()
        else:
            # update job status to failed
            self._file.update_status(-1, FileStatus.FAILURE, global_rv)

        # when the thread should exit, don't even bother decreasing njobs_running,
        # as it will should be set to 0 regardless, and you may end up with negative
        # njobs_running otherwise...
        if not self._should_exit:
            self._queue_manager._njobs_running -= 1

        return



from __future__ import annotations

import gi

gi.require_version("Gtk", "3.0")
from gi.repository import Gtk, GObject
import yaml

from typing import Dict, Any, Final
import logging

from .preferences import (
    Preference,
    BooleanPreference,
    ListPreference,
    DictPreference,
    StringPreference,
    Preferences,
)
from rfi_file_monitor.utils import EXPAND_AND_FILL, PREFERENCES_CONFIG_FILE

logger = logging.getLogger(__name__)


class PreferenceValueCellRenderer(Gtk.CellRenderer):
    @GObject.Property(type=str)
    def key(self) -> str:
        return self._key

    @key.setter
    def key(self, value: str):
        self._key = value
        self._set_renderer(value)

    def __init__(
        self,
        window: PreferencesWindow,
        settings: Dict[Preference, Any],
        list_store: Gtk.ListStore,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)

        self._settings = settings
        self._list_store = list_store
        # self._key: str = ""
        self._window: PreferencesWindow = window
        self._renderer: Gtk.CellRenderer = None

        # our renderers
        self._toggle_renderer = Gtk.CellRendererToggle(
            activatable=True, radio=False
        )
        self._combo_renderer = Gtk.CellRendererCombo(has_entry=False)
        self._text_renderer = Gtk.CellRendererText(editable=True)

        # our combo models
        self._combo_models: Final[Dict[Preference, Gtk.ListStore]] = dict()

        # connect signal handlers
        self._toggle_renderer.connect("toggled", self._toggle_cb)
        self._combo_renderer.connect("changed", self._changed_cb)
        self._text_renderer.connect("edited", self._edited_cb)

    def _get_key_from_model(self, path: str) -> str:
        return self._list_store[path][0]

    def _edited_cb(self, combo: Gtk.CellRendererText, path: str, new_text: str):
        key: str = self._get_key_from_model(path)
        pref: Preference = self._get_pref_for_key(key)
        self._settings[pref] = new_text

        # update config file
        self._window._update_config_file()

    def _changed_cb(
        self, combo: Gtk.CellRendererCombo, path: str, new_iter: Gtk.TreeIter
    ):
        key: str = self._get_key_from_model(path)
        pref: Preference = self._get_pref_for_key(key)
        store = self._combo_models[pref]

        new_value = store[new_iter][0]
        self._settings[pref] = new_value

        # update config file
        self._window._update_config_file()

    def _toggle_cb(self, renderer: Gtk.CellRendererToggle, path: str):
        key: str = self._get_key_from_model(path)
        pref: Preference = self._get_pref_for_key(key)
        self._settings[pref] = not self._settings[pref]

        # update config file
        self._window._update_config_file()

    def _get_pref_for_key(self, key) -> Preference:
        # given a key, get the corresponding Preference class
        for _pref in self._settings:
            if _pref.key == key:
                return _pref
        raise Exception(f"pref not found for key {key}")

    def _set_renderer(self, key: str):
        pref: Preference = self._get_pref_for_key(key)

        if isinstance(pref, BooleanPreference):
            self._renderer = self._toggle_renderer
            # the mode has to be set for both self and child!!!
            self.props.mode = Gtk.CellRendererMode.ACTIVATABLE
            self._renderer.props.mode = Gtk.CellRendererMode.ACTIVATABLE
            self._renderer.props.active = self._settings[pref]
            self._renderer.props.activatable = True
        elif isinstance(pref, ListPreference) or isinstance(
            pref, DictPreference
        ):
            self._renderer = self._combo_renderer
            self.props.mode = Gtk.CellRendererMode.EDITABLE
            current_value = self._settings[pref]

            if pref not in self._combo_models:
                # create new model
                store = Gtk.ListStore(str)
                for _val in pref.values:
                    store.append([_val])
                self._combo_models[pref] = store
            else:
                store = self._combo_models[pref]

            self._renderer.props.model = store
            self._renderer.props.text = current_value
            self._renderer.props.text_column = 0
            self._renderer.props.editable = True
            self._renderer.props.mode = Gtk.CellRendererMode.EDITABLE
        elif isinstance(pref, StringPreference):
            self._renderer = self._text_renderer
            self.props.mode = Gtk.CellRendererMode.EDITABLE
            current_value = self._settings[pref]
            self._renderer.props.mode = Gtk.CellRendererMode.EDITABLE
            self._renderer.props.editable = True
            self._renderer.props.text = current_value
        else:
            raise NotImplementedError

    # these methods define how the renderer should do its drawing.
    # we just need to redirect it to the appropriate child renderer.
    def do_activate(
        self, event, widget, path, background_area, cell_area, flags
    ):
        return type(self._renderer).do_activate(
            self._renderer,
            event,
            widget,
            path,
            background_area,
            cell_area,
            flags,
        )

    def do_get_aligned_area(self, widget, flags, cell_area):
        return type(self._renderer).do_get_aligned_area(
            self._renderer, widget, flags, cell_area
        )

    def do_get_preferred_height(self, widget):
        return type(self._renderer).do_get_preferred_height(
            self._renderer, widget
        )

    def do_get_preferred_height_for_width(self, widget, width):
        return type(self._renderer).do_get_preferred_height_for_width(
            self._renderer, widget, width
        )

    def do_get_preferred_width(self, widget):
        return type(self._renderer).do_get_preferred_width(
            self._renderer, widget
        )

    def do_get_preferred_width_for_height(self, widget, height):
        return type(self._renderer).do_get_preferred_width_for_height(
            self._renderer, widget, height
        )

    def do_get_request_mode(self):
        return type(self._renderer).do_get_request_mode(self._renderer)

    def do_get_size(self, widget, cell_area):
        return type(self._renderer).do_get_size(
            self._renderer, widget, cell_area
        )

    def do_render(self, cr, widget, background_area, cell_area, flags):
        type(self._renderer).do_render(
            self._renderer, cr, widget, background_area, cell_area, flags
        )

    def do_start_editing(
        self, event, widget, path, background_area, cell_area, flags
    ):
        return type(self._renderer).do_start_editing(
            self._renderer,
            event,
            widget,
            path,
            background_area,
            cell_area,
            flags,
        )


class PreferencesWindow(Gtk.Window):
    def __init__(self, prefs: Preferences, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.set_default_size(500, 500)
        self._prefs: Preferences = prefs

        grid = Gtk.Grid(**EXPAND_AND_FILL)
        self.add(grid)

        nb = Gtk.Notebook(**EXPAND_AND_FILL)
        grid.attach(nb, 0, 0, 1, 1)

        # Settings
        config_page = Gtk.Grid(
            **EXPAND_AND_FILL, border_width=5, row_spacing=5, column_spacing=5
        )
        nb.append_page(config_page, Gtk.Label("Settings"))

        frame_child = Gtk.Label(
            label="<b>Settings specific to this machine, usable by operations and engines</b>",
            use_markup=True,
            halign=Gtk.Align.FILL,
            valign=Gtk.Align.CENTER,
            hexpand=False,
            vexpand=False,
        )
        frame = Gtk.Frame(
            height_request=50,
            halign=Gtk.Align.FILL,
            valign=Gtk.Align.START,
            hexpand=True,
            vexpand=False,
        )
        frame.add(frame_child)
        config_page.attach(frame, 0, 0, 1, 1)

        store = Gtk.ListStore(str, str)
        for _pref in self._prefs.settings:
            store.append([_pref.key, _pref.description])

        sw = Gtk.ScrolledWindow(
            **EXPAND_AND_FILL, shadow_type=Gtk.ShadowType.IN
        )
        tv = Gtk.TreeView(model=store, tooltip_column=1, **EXPAND_AND_FILL)
        sw.add(tv)
        config_page.attach(sw, 0, 1, 1, 1)

        key_renderer = Gtk.CellRendererText()
        value_renderer = PreferenceValueCellRenderer(
            window=self, settings=self._prefs.settings, list_store=store
        )

        key_column = Gtk.TreeViewColumn("Key", key_renderer, text=0)
        value_column = Gtk.TreeViewColumn("Value", value_renderer, key=0)

        tv.append_column(key_column)
        tv.append_column(value_column)

        # Operations
        operations_page = Gtk.Grid(
            **EXPAND_AND_FILL, border_width=5, row_spacing=5, column_spacing=5
        )
        nb.append_page(operations_page, Gtk.Label("Operations"))

        frame_child = Gtk.Label(
            label="<b>Use this table to select which operations will be available.</b>",
            use_markup=True,
            halign=Gtk.Align.FILL,
            valign=Gtk.Align.CENTER,
            hexpand=False,
            vexpand=False,
        )
        frame = Gtk.Frame(
            height_request=50,
            halign=Gtk.Align.FILL,
            valign=Gtk.Align.START,
            hexpand=True,
            vexpand=False,
        )
        frame.add(frame_child)
        operations_page.attach(frame, 0, 0, 1, 1)

        self._operations_store = Gtk.ListStore(str, bool, object)
        for _op, _value in self._prefs.operations.items():
            self._operations_store.append([_op.NAME, _value, _op])

        sw = Gtk.ScrolledWindow(
            **EXPAND_AND_FILL, shadow_type=Gtk.ShadowType.IN
        )
        tv = Gtk.TreeView(model=self._operations_store, **EXPAND_AND_FILL)
        sw.add(tv)
        operations_page.attach(sw, 0, 1, 1, 1)

        key_renderer = Gtk.CellRendererText()
        value_renderer = Gtk.CellRendererToggle()
        value_renderer.connect("toggled", self._operation_toggled)

        key_column = Gtk.TreeViewColumn("Operation Name", key_renderer, text=0)
        value_column = Gtk.TreeViewColumn("Enabled", value_renderer, active=1)

        tv.append_column(key_column)
        tv.append_column(value_column)

        # Engines
        engines_page = Gtk.Grid(
            **EXPAND_AND_FILL, border_width=5, row_spacing=5, column_spacing=5
        )
        nb.append_page(engines_page, Gtk.Label("Engines"))

        frame_child = Gtk.Label(
            label="<b>Use this table to select which engines will be available.</b>",
            use_markup=True,
            halign=Gtk.Align.FILL,
            valign=Gtk.Align.CENTER,
            hexpand=False,
            vexpand=False,
        )
        frame = Gtk.Frame(
            height_request=50,
            halign=Gtk.Align.FILL,
            valign=Gtk.Align.START,
            hexpand=True,
            vexpand=False,
        )
        frame.add(frame_child)
        engines_page.attach(frame, 0, 0, 1, 1)

        self._engines_store = Gtk.ListStore(str, bool, object)
        for _engine, _value in self._prefs.engines.items():
            self._engines_store.append([_engine.NAME, _value, _engine])

        sw = Gtk.ScrolledWindow(
            **EXPAND_AND_FILL, shadow_type=Gtk.ShadowType.IN
        )
        tv = Gtk.TreeView(model=self._engines_store, **EXPAND_AND_FILL)
        sw.add(tv)
        engines_page.attach(sw, 0, 1, 1, 1)

        key_renderer = Gtk.CellRendererText()
        value_renderer = Gtk.CellRendererToggle()
        value_renderer.connect("toggled", self._engine_toggled)

        key_column = Gtk.TreeViewColumn("Engine Name", key_renderer, text=0)
        value_column = Gtk.TreeViewColumn("Enabled", value_renderer, active=1)

        tv.append_column(key_column)
        tv.append_column(value_column)

        grid.show_all()

    def _operation_toggled(self, renderer: Gtk.CellRendererToggle, path: str):
        count_toggled = 0
        for _toggled in self._operations_store:
            if _toggled[1] is True:
                count_toggled += 1
        if count_toggled == 1 and self._operations_store[path][1] is True:
            dialog = Gtk.MessageDialog(
                transient_for=self,
                modal=True,
                destroy_with_parent=True,
                message_type=Gtk.MessageType.ERROR,
                buttons=Gtk.ButtonsType.CLOSE,
                text=f"At least 1 operation must be selected!",
            )
            dialog.run()
            dialog.destroy()
            return
        self._operations_store[path][1] = not self._operations_store[path][1]
        self._prefs.operations[
            self._operations_store[path][2]
        ] = self._operations_store[path][1]

        self._update_config_file()

    def _engine_toggled(self, renderer: Gtk.CellRendererToggle, path: str):
        count_toggled = 0
        for _toggled in self._engines_store:
            if _toggled[1] is True:
                count_toggled += 1
        if count_toggled == 1 and self._engines_store[path][1] is True:
            dialog = Gtk.MessageDialog(
                transient_for=self,
                modal=True,
                destroy_with_parent=True,
                message_type=Gtk.MessageType.ERROR,
                buttons=Gtk.ButtonsType.CLOSE,
                text=f"At least 1 engine must be selected!",
            )
            dialog.run()
            dialog.destroy()
            return
        self._engines_store[path][1] = not self._engines_store[path][1]
        self._prefs.engines[self._engines_store[path][2]] = self._engines_store[
            path
        ][1]

        self._update_config_file()

    def _update_config_file(self):
        # write prefs into dictionary format
        yaml_dict = dict(settings={}, operations={}, engines={})

        for _pref, _value in self._prefs.settings.items():
            yaml_dict["settings"][_pref.key] = _value

        for _op, _value in self._prefs.operations.items():
            yaml_dict["operations"][_op.NAME] = _value

        for _engine, _value in self._prefs.engines.items():
            yaml_dict["engines"][_engine.NAME] = _value

        try:
            # ensure parent directories of preferences file have been created
            PREFERENCES_CONFIG_FILE.parent.mkdir(
                mode=0o700, parents=True, exist_ok=True
            )

            # open for writing
            with PREFERENCES_CONFIG_FILE.open("w") as f:
                yaml.safe_dump(data=yaml_dict, stream=f)
        except Exception:
            logger.exception(
                f"Could not write to {str(PREFERENCES_CONFIG_FILE)}"
            )



import gi

gi.require_version("Gtk", "3.0")
from gi.repository import Gtk, Gio, GLib

from ..operation import Operation
from ..utils.exceptions import SkippedOperation
from ..file import File, RegularFile, Directory
from ..utils import get_md5
from ..utils.decorators import (
    with_pango_docs,
    supported_filetypes,
    add_directory_support,
)

import logging
import tempfile
import os
from pathlib import Path
from threading import current_thread

logger = logging.getLogger(__name__)


@with_pango_docs(filename="local_copier.pango")
@supported_filetypes(filetypes=(RegularFile, Directory))
class LocalCopierOperation(Operation):

    NAME = "Local Copier"

    def __init__(self, *args, **kwargs):
        Operation.__init__(self, *args, **kwargs)
        grid = Gtk.Grid(
            border_width=5,
            row_spacing=5,
            column_spacing=5,
            halign=Gtk.Align.FILL,
            valign=Gtk.Align.CENTER,
            hexpand=True,
            vexpand=False,
        )
        self.add(grid)

        # boxes are needed for
        # destination folder
        # that's it??
        # a button to force creation of the destination folder is not necessary,
        # as we use a filechooserdialog to select it, and create it from there if necessary.

        label = Gtk.Label(
            label="Destination",
            halign=Gtk.Align.START,
            valign=Gtk.Align.CENTER,
            hexpand=False,
            vexpand=False,
        )
        grid.attach(label, 0, 0, 1, 1)

        directory_chooser_button = self.register_widget(
            Gtk.FileChooserButton(
                title="Select a directory to copy monitored files to",
                action=Gtk.FileChooserAction.SELECT_FOLDER,
                create_folders=True,
                halign=Gtk.Align.FILL,
                valign=Gtk.Align.CENTER,
                hexpand=True,
                vexpand=False,
            ),
            "destination_directory",
        )
        grid.attach(directory_chooser_button, 1, 0, 1, 1)

        grid.attach(
            Gtk.Separator(
                orientation=Gtk.Orientation.HORIZONTAL,
                halign=Gtk.Align.FILL,
                valign=Gtk.Align.CENTER,
                hexpand=True,
                vexpand=False,
            ),
            0,
            1,
            2,
            1,
        )

        overwrite_checkbutton = self.register_widget(
            Gtk.CheckButton(
                label="Overwrite existing files",
                active=True,
                halign=Gtk.Align.FILL,
                valign=Gtk.Align.CENTER,
                hexpand=True,
                vexpand=False,
            ),
            "overwrite",
        )
        grid.attach(overwrite_checkbutton, 0, 2, 2, 1)

    def preflight_check(self):
        # ensure destination is not None
        if self.params.destination_directory is None:
            raise ValueError("Destination folder cannot be empty")

        # destination directory cannot be monitored directory!!

        from ..engines.file_watchdog_engine import FileWatchdogEngine

        if isinstance(self.appwindow.active_engine, FileWatchdogEngine):

            if Path(self.params.destination_directory).samefile(
                self.appwindow.active_engine.params.monitored_directory
            ):
                raise ValueError(
                    "Destination folder cannot be the same as the monitored directory"
                )

            # when using recursive monitoring, the destination directory cannot be a subdirectory of the monitored directory
            if self.appwindow.active_engine.params.monitor_recursively:
                try:
                    Path(
                        self.params.destination_directory
                    ).resolve().relative_to(
                        Path(
                            self.appwindow.active_engine.params.monitored_directory
                        )
                    )
                except ValueError:
                    pass
                else:
                    raise ValueError(
                        "The destination directory cannot be a subdirectory of the monitored directory when monitoring recursively."
                    )

        with tempfile.NamedTemporaryFile(delete=False) as f:
            f.write(os.urandom(1024))
            tmpfile = f.name
        try:
            gtmpfile = Gio.File.new_for_path(tmpfile)
            destination_file = str(
                Path(self.params.destination_directory, Path(tmpfile).name)
            )
            gdestination_file = Gio.File.new_for_path(destination_file)
            if self.params.overwrite:
                self._flags = Gio.FileCopyFlags.OVERWRITE
            else:
                self._flags = Gio.FileCopyFlags.NONE
            gtmpfile.copy(gdestination_file, self._flags)
        except GLib.Error:
            logger.exception(
                f"Error copying {tmpfile} to {self.params.destination_directory}"
            )
            raise
        else:
            # delete copied file
            os.unlink(destination_file)
        finally:
            # remove temporary file
            os.unlink(tmpfile)

    def _attach_metadata(self, file: File, destination_file: Path):
        file.operation_metadata[self.index] = {
            "local copy path": destination_file
        }

    @add_directory_support
    def run(self, file: File):
        try:
            gsource_file = Gio.File.new_for_path(file.filename)
            destination_file = Path(
                self.params.destination_directory, *file.relative_filename.parts
            )

            # skip if the file has been copied already
            if (
                destination_file.exists()
                and Path(file.filename).stat().st_size
                == destination_file.stat().st_size
            ):
                # calculate checksums
                local_md5 = get_md5(file.filename)
                copy_md5 = get_md5(destination_file)

                if local_md5 == copy_md5:
                    # add destination path to metadata
                    self._attach_metadata(file, destination_file)
                    raise SkippedOperation("File has been copied already")

            # make parent directories if necessary
            destination_file.parent.mkdir(parents=True, exist_ok=True)
            gdestination_file = Gio.File.new_for_path(str(destination_file))
            gsource_file.copy(
                gdestination_file,
                self._flags,
                file.cancellable,
                LocalCopyProgressPercentage(file, self),
            )
        except SkippedOperation:
            raise
        except Exception as e:
            logger.exception(f"LocalCopierOperation.run exception")
            return str(e)
        else:
            # add destination path to metadata
            self._attach_metadata(file, destination_file)
        return None


class LocalCopyProgressPercentage:
    def __init__(self, file: File, operation: Operation):
        self._file = file
        self._last_percentage = 0
        self._operation = operation

    def __call__(self, current_num_bytes, total_num_bytes):
        thread = current_thread()
        if thread.should_exit:
            self._file.cancellable.cancel()
            return
        percentage = (current_num_bytes / total_num_bytes) * 100
        if int(percentage) > self._last_percentage:
            self._last_percentage = int(percentage)
            self._file.update_progressbar(
                self._operation.index, self._last_percentage
            )



from __future__ import annotations

import gi

gi.require_version("Gtk", "3.0")
from gi.repository import Gtk, GLib, GObject

from typing import OrderedDict as OrderedDictType
from typing import Final, List, Union, Sequence, Optional
from collections import OrderedDict
from threading import RLock
import logging
import os
from time import time
from dataclasses import dataclass, astuple as dc_astuple

from .file import FileStatus, File
from .job import Job
from .utils.exceptions import AlreadyRunning, NotYetRunning
from .utils.widgetparams import WidgetParams

logger = logging.getLogger(__name__)


class QueueManager(WidgetParams, Gtk.Grid):
    MAX_JOBS = (
        len(getattr(os, "sched_getaffinity")(0))
        if hasattr(os, "sched_getaffinity")
        else os.cpu_count()
    )

    NAME = "Queue Manager"

    def __init__(self, appwindow):
        self._appwindow = appwindow
        self._running = False
        self._files_dict_lock = RLock()
        self._files_dict: OrderedDictType[str, File] = OrderedDict()
        self._jobs_list: Final[List[Job]] = list()
        self._njobs_running: int = 0

        kwargs = dict(
            halign=Gtk.Align.FILL,
            valign=Gtk.Align.FILL,
            hexpand=True,
            vexpand=True,
            border_width=5,
            row_spacing=5,
        )
        Gtk.Grid.__init__(self, **kwargs)
        WidgetParams.__init__(self)

        self.options_child_row_counter = 0

        # Promote created files to saved after # seconds
        created_status_promotion_grid = Gtk.Grid(
            halign=Gtk.Align.FILL,
            valign=Gtk.Align.CENTER,
            hexpand=True,
            vexpand=False,
            column_spacing=5,
        )

        self.attach(
            created_status_promotion_grid,
            0,
            self.options_child_row_counter,
            1,
            1,
        )
        self.options_child_row_counter += 1
        created_status_promotion_checkbutton = self.register_widget(
            Gtk.CheckButton(
                label="Promote files from 'Created' to 'Saved' after",
                halign=Gtk.Align.START,
                valign=Gtk.Align.CENTER,
                hexpand=False,
                vexpand=False,
            ),
            "created_status_promotion_active",
            desensitized=True,
        )
        created_status_promotion_grid.attach(
            created_status_promotion_checkbutton, 0, 0, 1, 1
        )
        created_status_promotion_spinbutton = self.register_widget(
            Gtk.SpinButton(
                adjustment=Gtk.Adjustment(
                    lower=1, upper=3600, value=5, page_size=0, step_increment=1
                ),
                value=5,
                update_policy=Gtk.SpinButtonUpdatePolicy.IF_VALID,
                numeric=True,
                climb_rate=5,
                halign=Gtk.Align.CENTER,
                valign=Gtk.Align.CENTER,
                hexpand=False,
                vexpand=False,
            ),
            "created_status_promotion_delay",
            desensitized=True,
        )
        created_status_promotion_grid.attach(
            created_status_promotion_spinbutton, 1, 0, 1, 1
        )
        created_status_promotion_grid.attach(
            Gtk.Label(label="seconds"), 2, 0, 1, 1
        )

        self._add_horizontal_separator()

        # Delay promoting saved files to queued for # seconds
        saved_status_promotion_grid = Gtk.Grid(
            halign=Gtk.Align.FILL,
            valign=Gtk.Align.CENTER,
            hexpand=True,
            vexpand=False,
            column_spacing=5,
        )

        self.attach(
            saved_status_promotion_grid, 0, self.options_child_row_counter, 1, 1
        )
        self.options_child_row_counter += 1
        label = Gtk.Label(
            label="Delay promoting files from 'Saved' to 'Queued' for",
            halign=Gtk.Align.START,
            valign=Gtk.Align.CENTER,
            hexpand=False,
            vexpand=False,
        )
        saved_status_promotion_grid.attach(label, 0, 0, 1, 1)
        saved_status_promotion_spinbutton = self.register_widget(
            Gtk.SpinButton(
                adjustment=Gtk.Adjustment(
                    lower=2, upper=3600, value=5, page_size=0, step_increment=1
                ),
                value=5,
                update_policy=Gtk.SpinButtonUpdatePolicy.IF_VALID,
                numeric=True,
                climb_rate=5,
                halign=Gtk.Align.CENTER,
                valign=Gtk.Align.CENTER,
                hexpand=False,
                vexpand=False,
            ),
            "saved_status_promotion_delay",
            desensitized=True,
        )
        saved_status_promotion_grid.attach(
            saved_status_promotion_spinbutton, 1, 0, 1, 1
        )
        saved_status_promotion_grid.attach(
            Gtk.Label(label="seconds"), 2, 0, 1, 1
        )

        self._add_horizontal_separator()

        # Set the max number of threads to be used for processing
        max_threads_grid = Gtk.Grid(
            halign=Gtk.Align.FILL,
            valign=Gtk.Align.CENTER,
            hexpand=True,
            vexpand=False,
            column_spacing=5,
        )
        self.attach(max_threads_grid, 0, self.options_child_row_counter, 1, 1)
        self.options_child_row_counter += 1
        max_threads_grid.attach(
            Gtk.Label(
                label="Maximum number of threads to use",
                halign=Gtk.Align.START,
                valign=Gtk.Align.CENTER,
                hexpand=False,
                vexpand=False,
            ),
            0,
            0,
            1,
            1,
        )

        max_threads_spinbutton = self.register_widget(
            Gtk.SpinButton(
                adjustment=Gtk.Adjustment(
                    lower=1,
                    upper=self.MAX_JOBS,
                    value=max(self.MAX_JOBS // 2, 1),
                    page_size=0,
                    step_increment=1,
                ),
                value=max(self.MAX_JOBS // 2, 1),
                update_policy=Gtk.SpinButtonUpdatePolicy.IF_VALID,
                numeric=True,
                climb_rate=1,
                halign=Gtk.Align.START,
                valign=Gtk.Align.CENTER,
                hexpand=False,
                vexpand=False,
            ),
            "max_threads",
            desensitized=True,
        )
        max_threads_grid.attach(max_threads_spinbutton, 1, 0, 1, 1)

        self._add_horizontal_separator()

        # Remove from list after n minutes
        remove_from_list_status_promotion_grid = Gtk.Grid(
            halign=Gtk.Align.FILL,
            valign=Gtk.Align.CENTER,
            hexpand=True,
            vexpand=False,
            column_spacing=5,
        )

        self.attach(
            remove_from_list_status_promotion_grid,
            0,
            self.options_child_row_counter,
            1,
            1,
        )
        self.options_child_row_counter += 1
        remove_from_list_status_promotion_checkbutton = self.register_widget(
            Gtk.CheckButton(
                label="Remove from table after",
                active=True,
                halign=Gtk.Align.START,
                valign=Gtk.Align.CENTER,
                hexpand=False,
                vexpand=False,
            ),
            "remove_from_list_status_promotion_active",
            desensitized=True,
        )
        remove_from_list_status_promotion_grid.attach(
            remove_from_list_status_promotion_checkbutton, 0, 0, 1, 1
        )
        remove_from_list_status_promotion_spinbutton = self.register_widget(
            Gtk.SpinButton(
                adjustment=Gtk.Adjustment(
                    lower=1,
                    upper=60 * 24 * 7,
                    value=60,
                    page_size=0,
                    step_increment=1,
                ),
                value=60,
                update_policy=Gtk.SpinButtonUpdatePolicy.IF_VALID,
                numeric=True,
                climb_rate=5,
                halign=Gtk.Align.CENTER,
                valign=Gtk.Align.CENTER,
                hexpand=False,
                vexpand=False,
            ),
            "remove_from_list_status_promotion_delay",
            desensitized=True,
        )
        remove_from_list_status_promotion_grid.attach(
            remove_from_list_status_promotion_spinbutton, 1, 0, 1, 1
        )
        remove_from_list_status_promotion_grid.attach(
            Gtk.Label(label="minutes"), 2, 0, 1, 1
        )

    def _add_horizontal_separator(self):
        self.attach(
            Gtk.Separator(
                orientation=Gtk.Orientation.HORIZONTAL,
                halign=Gtk.Align.FILL,
                valign=Gtk.Align.CENTER,
                hexpand=True,
                vexpand=True,
            ),
            0,
            self.options_child_row_counter,
            1,
            1,
        )
        self.options_child_row_counter += 1

    @GObject.Property(type=bool, default=False)
    def running(self):
        return self._running

    def _add_to_model(self, file: File):
        # add new entry to model
        outputrow = OutputRow(
            relative_filename=str(file.relative_filename),
            creation_timestamp=file.created,
            status=int(file.status),
            operation_name="All",
        )
        iter = self._appwindow._files_tree_model.append(
            parent=None, row=dc_astuple(outputrow)
        )
        _row_reference = Gtk.TreeRowReference.new(
            self._appwindow._files_tree_model,
            self._appwindow._files_tree_model.get_path(iter),
        )

        # create its children, one for each operation
        for _operation in self._appwindow._operations_box:
            outputrow = OutputRow(
                relative_filename="",
                creation_timestamp=0,
                status=int(FileStatus.QUEUED),
                operation_name=_operation.NAME,
            )
            self._appwindow._files_tree_model.append(
                parent=iter, row=dc_astuple(outputrow)
            )

        file.row_reference = _row_reference

    def add(self, file_or_files: Union[File, Sequence[File]]):
        """Add one or more new files to the queue. Call from the GUI thread!"""

        if not self._running:
            raise NotYetRunning(
                "The queue manager needs to be started before it can be stopped."
            )

        if isinstance(file_or_files, File):
            file_paths = [file_or_files]
        else:
            file_paths = list(file_or_files)

        with self._files_dict_lock:
            for _file in file_paths:
                if not isinstance(_file, File):
                    raise TypeError(f"{str(_file)} must be a File object")

                file_path = _file.filename

                if file_path in self._files_dict:
                    logger.info(
                        f"{file_path} has been recreated! Calling saved..."
                    )
                    self.saved(file_path)
                    continue

                if _file.status == FileStatus.CREATED:
                    logger.info(f"New file {file_path} created")
                elif _file.status == FileStatus.SAVED:
                    logger.info(f"Adding existing file {file_path}")
                    _file.saved = time()
                else:
                    raise NotImplementedError(
                        "Newly created files must have CREATED or SAVED as status!"
                    )

                self._add_to_model(_file)

                self._files_dict[file_path] = _file

    def saved(self, file_path: Union[str, Sequence[str]]):
        """Call when the engine detected that the file(s) have been saved (again).vMust be called from the GUI thread!"""

        if not self._running:
            raise NotYetRunning(
                "The queue manager needs to be started before it can be stopped."
            )

        if isinstance(file_path, str):
            file_paths = [file_path]
        else:
            file_paths = list(file_path)

        with self._files_dict_lock:
            for file_path in file_paths:
                try:
                    file = self._files_dict[file_path]
                except KeyError:
                    logger.warning(
                        f"{file_path} has not been created yet! Ignoring..."
                    )
                    continue

                if file.status == FileStatus.SAVED:
                    # looks like this file has been saved again!
                    # update saved timestamp
                    logger.info(f"File {file_path} has been saved again")
                    file.saved = time()
                elif file.status == FileStatus.CREATED:
                    logger.info(f"File {file_path} has been saved")
                    file.status = FileStatus.SAVED
                    file.saved = time()
                    path = file.row_reference.get_path()
                    self._appwindow._files_tree_model[path][2] = int(
                        FileStatus.SAVED
                    )
                elif file.status == FileStatus.QUEUED:
                    # file hasn't been processed yet, so it's safe to demote it to SAVED
                    logger.info(
                        f"File {file_path} has been saved again while queued"
                    )
                    file.status = FileStatus.SAVED
                    path = file.row_reference.get_path()
                    self._appwindow._files_tree_model[path][2] = int(
                        FileStatus.SAVED
                    )
                elif file.status in (
                    FileStatus.RUNNING,
                    FileStatus.SUCCESS,
                    FileStatus.FAILURE,
                ):
                    # file is currently being processed or has been processed -> mark it for being requeued
                    logger.info(
                        f"File {file_path} has been saved again while {str(file.status)}"
                    )
                    file.requeue = True
                elif file.status is FileStatus.REMOVED_FROM_LIST:
                    # file has already been removed from the list!!
                    logger.info(
                        f"File {file_path} was removed from list but is now back!"
                    )
                    file.requeue = True
                    self._add_to_model(file)
                else:
                    logger.info(
                        f"File {file_path} has been saved again after it was queued for processing!!"
                    )

    def start(self):
        if self._running:
            raise AlreadyRunning(
                "The queue manager is already running. It needs to be stopped before it may be restarted"
            )

        self._running = True
        self._timeout_id = GLib.timeout_add_seconds(
            1, self._files_dict_timeout_cb, priority=GLib.PRIORITY_DEFAULT
        )
        self.notify("running")

    def stop(self):
        if not self._running:
            raise NotYetRunning(
                "The queue manager needs to be started before it can be stopped."
            )

        GLib.source_remove(self._timeout_id)
        with self._files_dict_lock:
            self._files_dict.clear()
            self._njobs_running = 0
        for job in self._jobs_list:
            job.should_exit = True
        self._jobs_list.clear()

        self._running = False
        self.notify("running")

    def _files_dict_timeout_cb(self, *user_data):
        """
        This function runs every second, and will take action based on the status of all files in the dict
        It runs in the GUI thread, so GUI updates are allowed here.
        """
        with self._files_dict_lock:
            status_counters = {
                FileStatus.CREATED: 0,
                FileStatus.SAVED: 0,
                FileStatus.QUEUED: 0,
                FileStatus.RUNNING: 0,
                FileStatus.SUCCESS: 0,
                FileStatus.FAILURE: 0,
                FileStatus.REMOVED_FROM_LIST: 0,
            }
            for _filename, _file in self._files_dict.items():
                if _file.status == FileStatus.CREATED:
                    if (
                        self.params.created_status_promotion_active
                        and (time() - _file.created)
                        > self.params.created_status_promotion_delay
                    ):
                        # promote to SAVED!
                        _file.status = FileStatus.SAVED
                        _file.saved = time()
                        path = _file.row_reference.get_path()
                        self._appwindow._files_tree_model[path][2] = int(
                            FileStatus.SAVED
                        )
                        logger.info(f"Promoting {_filename} to SAVED")

                elif _file.status == FileStatus.SAVED:
                    if (
                        time() - _file.saved
                    ) > self.params.saved_status_promotion_delay:
                        # queue the job
                        logger.info(
                            f"Adding {_filename} to queue for future processing"
                        )
                        _file.status = FileStatus.QUEUED
                        path = _file.row_reference.get_path()
                        self._appwindow._files_tree_model[path][2] = int(
                            _file.status
                        )

                elif _file.status == FileStatus.QUEUED:
                    if self._njobs_running < self.params.max_threads:
                        # try and launch a new job
                        logger.info(f"Launching queued job for {_filename}")
                        job = Job(self, _file)
                        self._jobs_list.append(job)
                        job.start()
                        self._njobs_running += 1

                elif _file.requeue and _file.status in (
                    FileStatus.SUCCESS,
                    FileStatus.FAILURE,
                    FileStatus.REMOVED_FROM_LIST,
                ):
                    # demote to saved so it gets requeued
                    _file.requeue = False
                    _file.status = FileStatus.SAVED
                    _file.saved = time()
                    _file.succeeded = 0
                    _file.operation_metadata.clear()
                    path = _file.row_reference.get_path()
                    iter = self._appwindow._files_tree_model.get_iter(path)
                    self._appwindow._files_tree_model[iter][2] = int(
                        FileStatus.SAVED
                    )
                    self._appwindow._files_tree_model[iter][4] = 0.0
                    self._appwindow._files_tree_model[iter][5] = "0.0 %"
                    self._appwindow._files_tree_model[iter][6] = None
                    self._appwindow._files_tree_model[iter][7] = ""

                    for child in self._appwindow._files_tree_model[
                        iter
                    ].iterchildren():
                        child[2] = int(FileStatus.QUEUED)
                        child[4] = 0.0
                        child[5] = "0.0 %"
                        child[6] = None
                        child[7] = ""

                    logger.info(f"Requeuing {_filename}")

                elif _file.status == FileStatus.SUCCESS:
                    if (
                        self.params.remove_from_list_status_promotion_active
                        and (
                            (time() - _file.succeeded)
                            > (
                                60
                                * self.params.remove_from_list_status_promotion_delay
                            )
                        )
                    ):

                        path = _file.row_reference.get_path()

                        # update status
                        _file.status = FileStatus.REMOVED_FROM_LIST
                        # remove from table
                        del self._appwindow._files_tree_model[path]

                status_counters[_file.status] += 1

            # update status bar
            self._appwindow._status_grid.get_child_at(
                0, 0
            ).props.label = f"Total: {len(self._files_dict)}"
            for _status, _counter in status_counters.items():
                self._appwindow._status_grid.get_child_at(
                    int(_status), 0
                ).props.label = f"{str(_status)}: {_counter}"

        return GLib.SOURCE_CONTINUE


@dataclass
class OutputRow:
    relative_filename: str
    creation_timestamp: float
    status: int
    operation_name: str
    operation_progress: float = 0.0
    operation_progress_str: str = "0.0 %"
    background_color: Optional[str] = None
    error_message: str = ""
